import numpy as np
from scipy.stats import logistic
from scipy.interpolate import interp1d
from scipy import integrate
import plasmapy.formulary as pp_form
import astropy.units as units
from .constants import *
from .daligault_gamma import DaligaultGamma
from .daligault_gamma import get_mu
from .fermi_gas_dos import FermiGasDOS
from .fermi_gas_params import FGParams
import warnings

warnings.filterwarnings("ignore")


class DaligaultAvgGamma:
    """
    class DaligaultAvgGamma calculates the energy-averaged scattering rate using scattering rates predicted by Eq. 4 in the 2017 Daligault PRL paper
    """

    def __init__(
        self,
        Te,
        N,
        V,
        theta_cutoff=1,
        q_kF_list=np.linspace(1e-3, 10, 1000),
        E=np.linspace(1e-6, 50, 100),
    ) -> None:
        """Function returns below parameters

        Args:
            Te (float): electronic temperature in eV
            N (int): Number of electrons in cell
            V (float): Volume of cell in Angstrom cubed (A^3)
            n (float): carrier density of material (1/m^3)
            theta (float): degeneracy parameter
            theta_cutoff (float): Te (eV) cutoff in which we go from quantum to classical chemical potential
            mu (float): chemical potential
            q_kF_list (array): array of q/kF points used for v(q) and G(q/kF), defaults to np.linspace(1e-3, 10, 1000) 1/m
            E (array): Energy array in eV from -500 to 500 eV
            E_gamma (2D array): Energy-mu (eV) and Daligault scattering rate (1/fs)
        """
        self.E = E
        self.Te = Te
        self.V = V
        self.N = N
        self.n = N / (V * angstrom**3)
        self.theta = pp_form.quantum_theta(self.Te * units.eV, self.n).value
        self.theta_cutoff = theta_cutoff
        # * get same mu value used in the scattering rate evaluation
        # TODO: clean this up a bit and unify in all scripts in the package
        self.mu = get_mu(self.n, self.theta, self.theta_cutoff) * J_to_eV  # eV
        self.q_kF_list = q_kF_list

        daligault_data = DaligaultGamma(self.E, self.Te, self.N, self.V)
        self.E_gamma = daligault_data.get_gamma()

    # get FG dos and then interpolates for data matching
    # The chemical potential should always be 0 eV for the scattering rates that are going to be used with this dos
    def get_fg_dos(self):
        # need this so that there is no interpolation out of range issue when shifting by mu
        E_range = np.linspace(-500, 5000, 1000000)
        fg_dos = FermiGasDOS(E_range, self.V)
        interp_fg_dos = interp1d(
            E_range - self.mu,
            fg_dos.dos,
            fill_value=(0, np.nan),
            bounds_error=False,
        )
        # mu is set to 0 eV
        return interp_fg_dos(self.E_gamma[:, 0])

    # FD distribution
    def get_occ(self):
        #! chemical potential should always be set to zero!
        return logistic.sf(self.E_gamma[:, 0], 0, self.Te)

    def daligault_gamma_occ_dos(self):
        E_gamma = self.E_gamma
        E = self.E_gamma[:, 0]
        gamma = self.E_gamma[:, 1]
        occ = DaligaultAvgGamma.get_occ(self)
        dos = DaligaultAvgGamma.get_fg_dos(self)
        return E, gamma, occ, dos

    def avg_daligault_gamma(self):
        E, gamma, occ, dos = DaligaultAvgGamma.daligault_gamma_occ_dos(self)
        avg_gamma = integrate.simpson(gamma * occ * (1 - occ) * dos, E)
        return avg_gamma



import numpy as np
from .constants import *
import warnings

warnings.filterwarnings("ignore")


class GWGamma:
    """
    class GWGamma generates all necessary parameters and outputs for analysis
    """

    def __init__(self, file, n=-1, core_E=-100000) -> None:
        """Function returns all parameters

        Args:
            file (string): the path to the OUTCAR file
            n (int): the scf step
            outcar (list): the OUTCAR file written to a list
            nkpts (int): the number of symmetry reduced KPOINTS
            algo (string): GW calculation algorithm
            Ef (float): Fermi energy of calculation, treated as the chemical potential
            core_E (float): cutoff (below) energy in eV that gets rid of core states - especially useful for integration
        """
        self.file = file
        self.n = n
        self.outcar = [line for line in open(self.file, "r")]
        self.nkpts = int(
            [line for line in self.outcar if "NKPTS" in line][0].split()[3]
        )
        self.algo = [line for line in self.outcar if "ALGO" in line][0].split()[-1]
        self.Ef = float(
            [line for line in self.outcar if "E-fermi" in line][0].split()[-1]
        )
        self.core_E = core_E  # eV

    def get_finite_Te_self_energy(self):
        """Function gets the energy and e-e scattering rates from the GW calculation
        Returns:
            E_gamma_occ (2D array - nkpts*nbands x 3): GW data of KS energies - Ef, scattering rate, and occupation values (0 to 1) sorted in ascending order of KS energies

        """

        ncol = 11
        head_string = "QP shifts evaluated in"
        tail_string = "Energies using frozen"
        header_num = [
            line_num for line_num, line in enumerate(self.outcar) if head_string in line
        ]
        tail_num = [
            line_num for line_num, line in enumerate(self.outcar) if tail_string in line
        ]
        # For non-single shot calculations
        if self.algo.lower() == "gwr" or self.algo.lower() == "scgwr":
            tail_string_gwr = "ROTSIGMA"
            tail_num_gw = [
                line_num
                for line_num, line in enumerate(self.outcar)
                if tail_string_gwr in line
            ]
            tail_num = [tail_num[0]] + tail_num_gw

        def get_clean_data(data):
            return (
                np.array(
                    [
                        [
                            x.split()
                            for x in [x.strip() for x in data if len(x.split()) == ncol]
                        ]
                    ]
                )
                .flatten()
                .astype(float)
            )

        finite_data = [
            get_clean_data(self.outcar[head:tail])
            for head, tail in list(zip(header_num, tail_num))
        ]

        nbands = int(len(finite_data[0]) / (self.nkpts * ncol))

        KS_energies = np.array(
            [
                x[1::ncol].reshape(
                    self.nkpts, nbands, int(len(x) / self.nkpts / nbands / ncol)
                )
                for x in finite_data
            ]
        ).flatten()
        occ = (
            np.array(
                [
                    x[7::ncol].reshape(
                        self.nkpts, nbands, int(len(x) / self.nkpts / nbands / ncol)
                    )
                    for x in finite_data
                ]
            ).flatten()
            / 2
        )
        gamma = (
            -2
            / hbar_eVfs
            * np.array(
                [
                    x[8::ncol].reshape(
                        self.nkpts, nbands, int(len(x) / self.nkpts / nbands / ncol)
                    )
                    for x in finite_data
                ]
            ).flatten()
        )
        # sort in order of KS-energies
        E_gamma_occ = np.array(
            sorted(np.array([KS_energies - self.Ef, gamma, occ]).T, key=lambda x: x[0])
        )

        # get rid of any core states
        E_gamma_occ = np.array([x for x in E_gamma_occ if x[0] >= self.core_E])

        return E_gamma_occ



import numpy as np
import warnings
from py4vasp import Calculation
from scipy.stats import logistic
from scipy import integrate

warnings.filterwarnings("ignore")


def func_format_dos(file, core_E):
    E_dos_array = np.array(
        [
            Calculation.from_file(f"{file}").dos.to_dict()["energies"],
            Calculation.from_file(f"{file}").dos.to_dict()["total"],
        ]
    ).T
    E_dos_array = np.array([x for x in E_dos_array if x[0] >= core_E]).T
    return E_dos_array[0], E_dos_array[1]


# binary search function
def func_binary_search(Te, N, E, dos, bs_E):
    bs_N = [integrate.simpson(logistic.sf(E, Ef, Te) * dos, E) for Ef in bs_E]
    if np.round(bs_N[1], 6) == N:
        return bs_E[1], bs_N[1]
    elif bs_N[1] > N:
        return func_binary_search(Te, N, E, dos, [bs_E[0], np.mean(bs_E[:2]), bs_E[1]])
    elif bs_N[1] < N:
        return func_binary_search(Te, N, E, dos, [bs_E[1], np.mean(bs_E[1:]), bs_E[2]])


# integrated DOS*occ to find the chemical potential for given Te
def func_integrate_for_mu(Te, N, E, dos):
    Ef, Ni = func_binary_search(
        Te, N, E, dos, [np.min(E), np.mean([np.min(E), np.max(E)]), np.max(E)]
    )
    return Ef, Ni


class VaspDOS:
    """
    class VaspDOS gets the energy and dos from a typical vasp DOS calc and also finds the chemical potential based on carrier conservation by integrating the occupations and dos over energy
    """

    def __init__(self, file, Te, N, core_E=-50) -> None:
        """Function returns all E and DOS

        Args:
            file (string): path+file to dos h5 file for py4vasp to post-processes
            Te (float): electronic temperature in eV
            E (array): Energy of VASP DOS from VASP
            N (int): Number of electrons in the calculation
            core_E (float): The energy which below is assumed to be only core contributions and excluded from DOS
            dos (array): Density of states calculated by VASP
            mu (float): The chemical potential found by integrating occupations and dos over energy, while constraining number of carriers
            Ni (float): The number of carriers found from finding the chemical potential, should be VERY close to N

        """
        self.file = file
        self.Te = Te
        self.N = N
        self.core_E = core_E
        self.E, self.dos = func_format_dos(self.file, self.core_E)
        self.mu, self.Ni = func_integrate_for_mu(self.Te, self.N, self.E, self.dos)



from scipy import constants


hbar_eVfs = (
    constants.hbar
    / constants.physical_constants["electron volt-joule relationship"][0]
    * 1e15
)  # eV/fs
J_to_eV = 1 / constants.physical_constants["electron volt-joule relationship"][0]
eV_to_au = 1 / constants.physical_constants["Hartree energy in eV"][0]
K_to_eV = 1 / constants.physical_constants["electron volt-kelvin relationship"][0]
eV_to_K = 1 / K_to_eV
eV_to_au = 1 / constants.physical_constants["Hartree energy in eV"][0]
# From KGCollection.cpp
hbar_Js = constants.hbar
m_e = constants.m_e
angstrom = constants.angstrom
charge_e = constants.elementary_charge
eV = constants.electron_volt
ohm_to_cgs = 0.11111111111111e-11
m_to_cgs = 1e2
hbar_eVs = (
    constants.hbar / constants.physical_constants["electron volt-joule relationship"][0]
)
c_cgs = 2.99792458e10
m_to_a0 = 1 / constants.physical_constants["Bohr radius"][0]
kB = constants.Boltzmann



from setuptools import setup, find_packages
from os import path

working_directory = path.abspath(path.dirname(__file__))

with open(path.join(working_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="qphGamma",  # name of package dir below module
    version="0.1.13",
    author="Brian Robinson",
    author_email="b.p.robinson102@gmail.com",
    description="A package for finding quasiparticle and quasihole scattering rates",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/brobinson10296/qphGamma",
    download_url="https://github.com/brobinson10296/qphGamma/archive/main.zip",
    keywords=[
        "first principles",
        "VASP",
        "Fermi gas",
        "Fermi liquid",
        "e-e scattering",
    ],
    packages=find_packages(where="./src"),
    package_dir={"": "src"},
    install_requires=["numpy", "pandas", "scipy", "py4vasp", "plasmapy"],
)
